/*
Murasaki - multiple genome global alignment program
Copyright (C) 2006-2007 Keio University
(Yasunori Osana) <comp@bio.keio.ac.jp> (2006)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "dbhash.h"
#include <set>

using namespace std;

// ----------------------------------------------------------------------
// MySQL Hash

#ifdef MURASAKI_MYSQL

void MySQLHash::createIndex(){
  if( !indexed ){
    mysql_query(fasta, "COMMIT");

    //  cout << "Sorting table..." << endl;
    mysql_query(fasta, "DROP TABLE murasakitemp");
    if(mysql_query(fasta, "CREATE TABLE murasakitemp (hashkey bigint, seq int, pos bigint) AS SELECT * FROM murasaki ORDER BY hashkey") != 0){
      cerr << "can't create sorted table" << endl;
      exit(-1);
    }

    mysql_query(fasta, "DROP TABLE murasaki");
    if(mysql_query(fasta, "RENAME TABLE murasakitemp TO murasaki") != 0){
      cerr << "can't rename sorted table" << endl;
      exit(-1);
    }

    //   cout << "Creating index..." << endl;
    indexed = true;
    mysql_query(fasta, "CREATE INDEX keyindex ON murasaki (hashkey)");
    //   cout << "Done with indexing..." << endl;

    mysql_query(fasta, "BEGIN");
  }
}

void MySQLHash::lookup(HashKey key,const HashVal &a,LocList &locList){
  vector<HashVal> v;

  createIndex();

  select_key = key;
  mysql_stmt_execute(stmt_select);
  //    mysql_stmt_bind_result(stmt_select, bind_selectr);

  if(mysql_stmt_store_result(stmt_select)){
    cerr << "mysql_stmt_store_result(): failed" << endl;
    exit(-1);
  }

  while(!mysql_stmt_fetch(stmt_select)){
    HashVal h;
    h.seqno = selectr_seq;
    h.pos   = selectr_pos;
    
    v.push_back(h);

    if((HashKey)selectr_key != key){
      cerr << "incorrect result from query?  requested=" << key 
	   << ", given=" << selectr_key << "," << selectr_seq <<"," << selectr_pos << endl;
      exit(-1);
    }
  }
  mysql_stmt_free_result(stmt_select);

  for(unsigned int i=0; i<v.size(); i++)
    //  if(a==v[i]) locList[v[i].seqId()].push_back(v[i]);
    if(!(a < v[i]) && !(v[i] < a)) locList[v[i].seqId()].push_back(v[i]);

}


void MySQLHash::lookup(HashKey key,LocList &locList){
  vector<HashVal> v;

  createIndex();

  select_key = key;
  if(mysql_stmt_execute(stmt_select)){
    cerr << "mysql_stmt_execute() failed for MySQLHash::lookup" <<endl;
    exit(-1);
  }
  //  mysql_stmt_bind_result(stmt_select, bind_selectr);
  mysql_stmt_store_result(stmt_select);

  while(!mysql_stmt_fetch(stmt_select)){
    HashVal h;
    h.seqno = selectr_seq;
    h.pos   = selectr_pos;
    
    v.push_back(h);

    if((HashKey)selectr_key != key){
      cerr << "incorrect result from query?  requested=" << key 
	   << ", given=" << selectr_key << "," << selectr_seq <<"," << selectr_pos << endl;
      exit(-1);
    }
  }
  mysql_stmt_free_result(stmt_select);

  for(unsigned int i=0; i<v.size(); i++)
    locList[v[i].seqId()].push_back(v[i]);
}

word MySQLHash::sizeAt(const HashKey key){
  int c = 0;

  createIndex();

  select_key = key;
  mysql_stmt_execute(stmt_select);
  mysql_stmt_store_result(stmt_select);
  c = mysql_stmt_num_rows(stmt_select);

  mysql_stmt_free_result(stmt_select);

  return c;
}

bool MySQLHash::emptyAt(const HashKey key){
  return (sizeAt(key) == 0);
}


void MySQLHash::add(const HashKey key, const HashVal &a){

  insert_key = key;
  insert_seq = (int)a.seqno;
  insert_pos = a.pos;

  //  cout << insert_key << "," << insert_seq << "," << insert_pos << endl;

  if (mysql_stmt_execute(stmt_insert)){
    fprintf(stderr, "MySQLHash::add() mysql_stmt_execute(), 1 failed\n");
    fprintf(stderr, " %s\n", mysql_stmt_error(stmt_insert));
    exit(0);
  }
}

void MySQLHash::clear(){

  mysql_query(fasta, "COMMIT");

  // Prepare tables
  mysql_query(fasta, "DROP TABLE murasaki");
  
  if(mysql_query(fasta,
		 "CREATE TABLE murasaki (hashkey bigint, seq int, pos bigint)") !=0){
    cerr << "couldn't create table!" << endl;
    exit(-1);
  } 

  // Prepare statements
  insert_SQL=strdup("INSERT INTO murasaki (hashkey,seq,pos) VALUES (?,?,?)");

  stmt_insert = mysql_stmt_init(fasta);
  mysql_stmt_prepare(stmt_insert, insert_SQL, strlen(insert_SQL));

  if( !stmt_insert ){
    cerr << "mysql_prepare(): failed." << endl;
    exit(-1);
  }

  if( mysql_stmt_param_count(stmt_insert)!=3 ){
    cerr << "mysql_prepare(): incorrect result from mysql_param_count(): "
         << mysql_stmt_param_count(stmt_insert)
	 << endl;
    exit(-1);
  }

  memset(bind_insert,  0, sizeof(bind_insert));

  bind_insert[0].buffer_type = MYSQL_TYPE_LONGLONG;
  bind_insert[1].buffer_type = MYSQL_TYPE_LONG;
  bind_insert[2].buffer_type = MYSQL_TYPE_LONGLONG;

  bind_select[0].buffer_type = MYSQL_TYPE_LONGLONG;

  bind_insert[0].buffer = &insert_key;
  bind_insert[1].buffer = &insert_seq;
  bind_insert[2].buffer = &insert_pos;

  bind_insert[0].buffer_length = sizeof(insert_key);
  bind_insert[1].buffer_length = sizeof(insert_seq);
  bind_insert[2].buffer_length = sizeof(insert_pos);

  bind_insert[0].is_null = 0;
  bind_insert[1].is_null = 0;
  bind_insert[2].is_null = 0;

  bind_insert[0].length = 0;
  bind_insert[1].length = 0;
  bind_insert[2].length = 0;

  mysql_stmt_bind_param(stmt_insert, bind_insert);


  ///

  select_SQL=strdup("SELECT * FROM murasaki where hashkey=?");
  // select_SQL=strdup("SELECT * FROM murasaki USE INDEX (keyindex) WHERE hashkey=?");
  stmt_select = mysql_stmt_init(fasta);
  mysql_stmt_prepare(stmt_select, select_SQL, strlen(select_SQL));

  if( !stmt_select ){
    cerr << "createIndex(): failed." << endl;
    exit(-1);
  }

  if( mysql_stmt_param_count(stmt_select)!=1 ){
    cerr << "createIndex(): incorrect result from mysql_param_count()"
	 << endl;
    exit(-1);
  }

  memset(bind_select,  0, sizeof(bind_select));

  bind_select[0].buffer_type = MYSQL_TYPE_LONGLONG;
  bind_select[0].buffer = (char*)&select_key;
  bind_select[0].buffer_length = sizeof(long long int);
  bind_select[0].is_null = 0;
  bind_select[0].length = 0;

  if(mysql_stmt_bind_param(stmt_select, bind_select)){
    cerr << "mysql_stmt_bind_param() for bind_select failed" << endl;
    exit(-1);
  }

  res_select = mysql_stmt_result_metadata(stmt_select);
  if(mysql_num_fields(res_select) != 3){
    cerr << "incorrect # of columns in meta result" << endl;
    exit(-1);
  }

  ///


  memset(bind_selectr, 0, sizeof(bind_selectr));

  bind_selectr[0].buffer_type = MYSQL_TYPE_LONGLONG;
  bind_selectr[1].buffer_type = MYSQL_TYPE_LONG;
  bind_selectr[2].buffer_type = MYSQL_TYPE_LONGLONG;

  bind_selectr[0].buffer = (char *)&selectr_key;
  bind_selectr[1].buffer = (char *)&selectr_seq;
  bind_selectr[2].buffer = (char *)&selectr_pos;

  bind_selectr[0].is_null = &selectr_nul[0];
  bind_selectr[1].is_null = &selectr_nul[1];
  bind_selectr[2].is_null = &selectr_nul[2];

  bind_selectr[0].length = &selectr_len[0];
  bind_selectr[1].length = &selectr_len[1];
  bind_selectr[2].length = &selectr_len[2];

  bind_selectr[0].error = &selectr_err[0];
  bind_selectr[1].error = &selectr_err[1];
  bind_selectr[2].error = &selectr_err[2];


  if(mysql_stmt_bind_result(stmt_select, bind_selectr)){
    cerr << "mysql_bind_result() failed." << endl;
    exit(-1);
  }

  ///

  // Start transaction
  mysql_query(fasta, "BEGIN");
}

MySQLHash::MySQLHash(BitSequence *pat) :  Hash(pat)  //holy crap C++ is cool
{
  // FIXME: check error or not

  indexed = false;

  fasta = mysql_init(NULL);
  // FIXME: the user/password
  if (mysql_real_connect(fasta, "localhost", "root", 
			 "debugordie" ,"murasaki", 0,NULL,0) == NULL){
    cerr << "couldn't connect to database" << endl;
    exit(-1);
  }

  clear();
}

MySQLHash::~MySQLHash(){
  assert(cout << "MySQLHash: closing database..." << endl);

  mysql_query(fasta, "COMMIT");
  mysql_close(fasta);
}

#endif

// ----------------------------------------------------------------------
// SQLite Hash

#ifdef MURASAKI_SQLITE3

static int sqlitehash_cb_lookup(void *v, int argc, char **argv, char **azColName){
  HashVal h;

  h.seqno = strtoul(argv[1], (char **)NULL, 10);
  h.pos   = atoi(argv[2]);

  (*(vector<HashVal>*)v).push_back(h);

  return 0;
}

void SQLiteHash::createIndex(){
  char *errmsg;

  if( !indexed ){

    cout << "Creating index..." << endl;
    indexed = true;
    //    (void)sqlite3_exec(fasta, "COMMIT", NULL, NULL, &errmsg);

    //    if(sqlite3_exec(fasta, "PRAGMA cache_size = 1000000",
    //              NULL, NULL, &errmsg)!=SQLITE_OK){
    //      fprintf(stderr, "Can't set cachesize: %s\n", sqlite3_errmsg(fasta));
    //    }

    if(sqlite3_exec(fasta, "CREATE INDEX keyindex ON murasaki (hashkey)",
                    NULL, NULL, &errmsg)!=SQLITE_OK){
      fprintf(stderr, "Can't create index: %s\n", sqlite3_errmsg(fasta));
    }

    if(sqlite3_exec(fasta, "CREATE INDEX keyindexlast ON murasakilast (hashkey)",
                    NULL, NULL, &errmsg)!=SQLITE_OK){
      fprintf(stderr, "Can't create index: %s\n", sqlite3_errmsg(fasta));
    }

    //    (void)sqlite3_exec(fasta, "BEGIN", NULL, NULL, &errmsg);
  }
}

void SQLiteHash::lookup(HashKey key,const HashVal &a,LocList &locList){
  char *sql, *errmsg;
  vector<HashVal> v;

  createIndex();

  sql = sqlite3_mprintf("SELECT * FROM murasaki where hashkey=%lu", key);
  (void)sqlite3_exec(fasta, sql, sqlitehash_cb_lookup, (void *)&v, &errmsg);
  free(sql);

  sql = sqlite3_mprintf("SELECT * FROM murasakilast where hashkey=%lu", key);
  (void)sqlite3_exec(fasta, sql, sqlitehash_cb_lookup, (void *)&v, &errmsg);
  free(sql);

  for(unsigned int i=0; i<v.size(); i++)
    if(!(a < v[i]) && !(v[i] < a)) locList[v[i].seqId()].push_back(v[i]);

}


void SQLiteHash::lookup(HashKey key,LocList &locList){
  char *sql, *errmsg;
  vector<HashVal> v;

  createIndex();

  sql = sqlite3_mprintf("SELECT * FROM murasaki where hashkey=%lu", key);
  (void)sqlite3_exec(fasta, sql, sqlitehash_cb_lookup, (void *)&v, &errmsg);
  free(sql);

  sql = sqlite3_mprintf("SELECT * FROM murasakilast where hashkey=%lu", key);
  (void)sqlite3_exec(fasta, sql, sqlitehash_cb_lookup, (void *)&v, &errmsg);
  free(sql);

  for(unsigned int i=0; i<v.size(); i++)
    locList[v[i].seqId()].push_back(v[i]);
}

static int sqlitehash_count(void *i, int argc, char **argv, char **azColName){
  // callback for sizeAt()
  (*(int*)i)++;

  return 0;
}

word SQLiteHash::sizeAt(const HashKey key){
  char *sql, *errmsg;
  int c=0;

  createIndex();

  sql = sqlite3_mprintf("SELECT * FROM murasaki where hashkey=%lu", key);
  (void)sqlite3_exec(fasta, sql, sqlitehash_count, &c, &errmsg); 
  free(sql);

  if(indexed){
    sql = sqlite3_mprintf("SELECT * FROM murasakilast where hashkey=%lu", key);
    (void)sqlite3_exec(fasta, sql, sqlitehash_count, &c, &errmsg); 
    free(sql);
  }

  return c;
}

bool SQLiteHash::emptyAt(const HashKey key){
  return (sizeAt(key) == 0);
}


void SQLiteHash::add(const HashKey key, const HashVal &a){
  char *sql, *errmsg;
  if(!indexed)
    sql = sqlite3_mprintf("INSERT INTO murasaki (hashkey, seq, pos) VALUES (%lu, %u, %d)", 
			  key, a.seqno, a.pos);
  else 
    sql = sqlite3_mprintf("INSERT INTO murasakilast (hashkey, seq, pos) VALUES (%lu, %u, %d)", 
			  key, a.seqno, a.pos);
  //  cout << sql << endl;;
  (void)sqlite3_exec(fasta, sql, NULL, NULL, &errmsg); 
  free(sql);
}

void SQLiteHash::clear(){
  char *errmsg;

  sqlite3_exec(fasta, "COMMIT", NULL, NULL, &errmsg);

  if(sqlite3_exec(fasta, "DROP TABLE murasaki", NULL, NULL, &errmsg)
     != SQLITE_OK){
    fprintf(stderr, "Can't drop tables: %s\n", sqlite3_errmsg(fasta));
    // ignore the errors
    //    sqlite3_close(fasta);
    //    exit(1);
  }

  if(sqlite3_exec(fasta, "DROP TABLE murasakilast", NULL, NULL, &errmsg)
     != SQLITE_OK){
    fprintf(stderr, "Can't drop tables: %s\n", sqlite3_errmsg(fasta));
    // ignore the errors
    //    sqlite3_close(fasta);
    //    exit(1);
  }
 
  if(sqlite3_exec(fasta, 
		  "CREATE TABLE murasaki(hashkey bigint, seq int, pos bigint)",
		  NULL, NULL, &errmsg)!=SQLITE_OK){
    fprintf(stderr, "Can't create tables: %s\n", sqlite3_errmsg(fasta));
    sqlite3_close(fasta);
    exit(1);
  }

  if(sqlite3_exec(fasta, 
		  "CREATE TABLE murasakilast(hashkey bigint, seq int, pos bigint)",
		  NULL, NULL, &errmsg)!=SQLITE_OK){
    fprintf(stderr, "Can't create tables: %s\n", sqlite3_errmsg(fasta));
    sqlite3_close(fasta);
    exit(1);
  }

  if(sqlite3_exec(fasta, "BEGIN", NULL, NULL, &errmsg)!=SQLITE_OK)
    fprintf(stderr, "Can't begin transaction: %s\n", sqlite3_errmsg(fasta));

}


SQLiteHash::SQLiteHash(BitSequence *pat) :  Hash(pat)  //holy crap C++ is cool
{
  int ret;
  char *errmsg;

  indexed = false;

  ret = sqlite3_open("murasaki-sqlite.db", &fasta); 

  if( ret ){
    fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(fasta));
    sqlite3_close(fasta);
    exit(1);
  }

  if(sqlite3_exec(fasta, "DROP TABLE murasaki", NULL, NULL, &errmsg)
     != SQLITE_OK){
    fprintf(stderr, "Can't drop tables: %s\n", sqlite3_errmsg(fasta));
    // ignore the errors
    //    sqlite3_close(fasta);
    //    exit(1);
  }

  if(sqlite3_exec(fasta, "DROP TABLE murasakilast", NULL, NULL, &errmsg)
     != SQLITE_OK){
    fprintf(stderr, "Can't drop tables: %s\n", sqlite3_errmsg(fasta));
    // ignore the errors
    //    sqlite3_close(fasta);
    //    exit(1);
  }
 
  if(sqlite3_exec(fasta, 
		  "CREATE TABLE murasaki(hashkey bigint, seq int, pos bigint)",
		  NULL, NULL, &errmsg)!=SQLITE_OK){
    fprintf(stderr, "Can't create tables: %s\n", sqlite3_errmsg(fasta));
    sqlite3_close(fasta);
    exit(1);
  }

  if(sqlite3_exec(fasta, 
		  "CREATE TABLE murasakilast(hashkey bigint, seq int, pos bigint)",
		  NULL, NULL, &errmsg)!=SQLITE_OK){
    fprintf(stderr, "Can't create tables: %s\n", sqlite3_errmsg(fasta));
    sqlite3_close(fasta);
    exit(1);
  }

  if(sqlite3_exec(fasta, "BEGIN", NULL, NULL, &errmsg)!=SQLITE_OK)
    fprintf(stderr, "Can't begin transaction: %s\n", sqlite3_errmsg(fasta));
}

SQLiteHash::~SQLiteHash(){
  char *errmsg;
  
  assert(cout << "SQLiteHash: closing database..." << endl);

  (void)sqlite3_exec(fasta, "COMMIT", NULL, NULL, &errmsg); 
  sqlite3_close(fasta); 
}

#endif

// ----------------------------------------------------------------------
